#!/bin/sh
#script version
VERSION="2.3.2"

# Standard exit codes:
#  0:     Success, driver loaded
#  1:     An error occurred
#  2:     Build in command failure
#  126:   Cannot execute command
#  127:   Command not found
#  128:   Invalid exit argument
#  128+N: Terminated by signal N
#  !0:    Syntax error

# Custom exit codes
#  101:   Success, no adapter present: shutdown?
#  102:   Success, current driver in use: reboot?


SHUTDOWN="No"
REBOOT="No"


if [ "$1" = "auto" ]; then
    INTERACTIVE=0
else
    INTERACTIVE=1
fi


# Purpose: Determine if the machine needs to be shutdown to install the adapter
#          for the drivers that was just installed.  Check $SHUTDOWN
#
# Params:  $1- The driver message key, i.e. the driver name
#
function checkLoadStatus() {
    # oswrap.c 1 -> 37
    MISSING_ORIG_MSG="Driver not loading"

    # oswrap.c 38 -> current
    MISSING_CURR_MSG="Driver will not be loaded because no ATTO \w+ devices were found"

    # Some drivers printed this with a 0 before printing "not loading", so don't
    # assume this means something was found
    FOUND_MSG="Found ([0-9])+ adapters"


    # Find the last line that is one of the status messages above.  Look for the
    # failure messages too in case there is a success in the previous lines of
    # output
    STATUS_LINE=`dmesg | grep $1 | tail -n 10 | tac | grep -m 1 -E "$MISSING_ORIG_MSG|$MISSING_CURR_MSG|$FOUND_MSG"`


    # Capture the number of adapters, if possible.  NB the space around the capture
	# group, so it doesn't match e.g. 2 in esas2raid.
    ADAPTERS=`echo $STATUS_LINE | grep -o -E " ([0-9])+ "`

    if [ "$ADAPTERS" = "" -o "$ADAPTERS" = " 0 " ]; then
        SHUTDOWN="Yes"; # something other than the found count was printed
    fi
}

function pauseandexit {
    if [ "$INTERACTIVE" == "1" ]; then
        echo "Press any key to finish..."
        read -n 1 -s
        echo
    fi
    exit $1
}

function checkforcommand {
    COMMAND=`which $1`
    if [ "${COMMAND}" = "" ]; then
        echo
        echo "ERROR: Unable to find '$1'."
        echo
        echo "       You must have '$1' to install this driver."
        echo
        echo "       Please refer to your system documentation for"
        echo "       instructions on obtaining '$1'."
        echo
        pauseandexit 1
    fi
}
    

echo
echo "ATTO Technology, Inc."
echo "Linux Driver Install Script v${VERSION}"
echo

# Check for a space in the current path name because
# the kernel Makefile is too stupid to deal with that

SPACE_IN_PWD=`pwd | grep " "`

if [ "${SPACE_IN_PWD}" != "" ]; then
    echo "ERROR: You are attempting to install this driver from a directory"
    echo "       which contains a space character somewhere in the path name."
    echo
    echo "       The Linux kernel Makefile cannot handle such a directory."
    echo
    echo "       Please extract the driver package to a different directory"
    echo "       and try again."
    echo
    pauseandexit 1
fi

# Check for kernel source code availability

FULL_KERNEL_VERSION=`uname -r`
KDIR="/lib/modules/${FULL_KERNEL_VERSION}/build"

if [ ! -x ${KDIR} ]; then
    echo
    echo "ERROR: Unable to find the kernel header directory."
    echo
    echo "       The kernel headers were expected in"
    echo "       ${KDIR}"
    echo
    echo "       Please refer to your system documentation for"
    echo "       instructions on obtaining the kernel headers."
    echo "       This may be called the \"kernel-devel\" package."
    echo
    echo "       If the kernel headers are installed in a different"
    echo "       location, you can build the driver manually from the src/"
    echo "       directory by specifying the correct kernel header directory"
    echo "       on the make command line with the KDIR variable, eg:"
    echo
    echo "       # make clean install KDIR=/usr/src/kernel"
    echo
    pauseandexit 1
fi

# Check for required commands

checkforcommand gcc
checkforcommand make

# Figure out what sort of driver we are building

DRIVER_NAME=`ls src/*core_x86.o | cut -d_ -f1 | cut -d/ -f2`
MP_DRIVER_NAME="UNDEFINED"
CONFLICT_DRIVER_NAME="UNDEFINED"
STD_DRIVER_NAME=${DRIVER_NAME}
VNIC_DRIVER_NAME=""

if [ "${DRIVER_NAME}" = "celerityfc" ]; then
    PRETTY_NAME="Celerity FC"
elif [ "${DRIVER_NAME}" = "express2" ]; then
    PRETTY_NAME="EPCI U320"
    CONFLICT_DRIVER_NAME="mptspi"
elif [ "${DRIVER_NAME}" = "esasraid" ]; then
    PRETTY_NAME="ESAS RAID"
elif [ "${DRIVER_NAME}" = "esashba" ]; then
    PRETTY_NAME="ESAS HBA"
elif [ "${DRIVER_NAME}" = "celerity8fc" ]; then
    PRETTY_NAME="Celerity FC 8GB"
    MP_DRIVER_NAME="celerity8fcmp"
elif [ "${DRIVER_NAME}" = "esas2hba" ]; then
    PRETTY_NAME="ExpressSAS 6Gb HBA"
    CONFLICT_DRIVER_NAME="pm8001"
elif [ "${DRIVER_NAME}" = "celerity8fcmp" ]; then
    PRETTY_NAME="Celerity Multipathing FC 8GB"
    STD_DRIVER_NAME="celerity8fc"
    MP_DRIVER_NAME=${DRIVER_NAME}
elif [ "${DRIVER_NAME}" = "esas2raid" ]; then
    PRETTY_NAME="ESAS 6Gb RAID"
elif [ "${DRIVER_NAME}" = "fastframe" ]; then
    PRETTY_NAME="FastFrame FCoE"
    CONFLICT_DRIVER_NAME="ixgbe"
    VNIC_DRIVER_NAME="ffxgbe"
else
    echo
    echo "ERROR: Unable to detect driver name"
    echo
    echo "       This driver package may have been corrupted."
    echo
    echo "       Please re-download the package and try again."
    echo
    pauseandexit 1
fi

# If loading standard driver, check for MP version, and vice-versa
MODULE_ROOT=/lib/modules/${FULL_KERNEL_VERSION}/kernel/drivers

STD_DRIVER_INSTALLED=`find ${MODULE_ROOT} -name ${STD_DRIVER_NAME}.ko | grep -v ${MP_DRIVER_NAME}`
MP_DRIVER_INSTALLED=`find ${MODULE_ROOT} -name ${MP_DRIVER_NAME}.ko`
CONFLICT_DRIVER_INSTALLED=`find ${MODULE_ROOT} -name ${CONFLICT_DRIVER_NAME}.ko`


if [ "$DRIVER_NAME" == "$STD_DRIVER_NAME" ]; then
  if [ "$MP_DRIVER_INSTALLED" != "" ]; then
    # unload & uninstall mp driver
    echo "A conflict has been discovered with a preexisting ATTO multipathing"
    echo "driver for this device."
    echo
    if [ "$INTERACTIVE" == "1" ]; then
        echo "You may uninstall the ${MP_DRIVER_NAME} driver here, note that"
        echo "multipathing features will no longer be available."
        echo "The driver module will be preserved and can be reinstalled if the "
        echo "${DRIVER_NAME} driver is uninstalled using the uninstall script."
        echo
        echo "Would you like to unload and uninstall ${MP_DRIVER_NAME}?    [N/y]"
        read -s -n 1 RESPONSE
    else
        echo "Unloading and uninstalling ${MP_DRIVER_NAME}..."
        RESPONSE="y"
    fi
    if [ "${RESPONSE}" = "y" -o "${RESPONSE}" = "Y" ]; then
        # Unload the driver and rename the module so that it can possibly be reinstalled later
        rmmod ${MP_DRIVER_NAME}
        echo
        echo "${MP_DRIVER_NAME} unloaded."
        mv ${MP_DRIVER_INSTALLED} ${MP_DRIVER_INSTALLED/%.ko/.bkup}
        echo "${MP_DRIVER_NAME} backed up."
        echo
    fi
  fi
fi


if [ "$DRIVER_NAME" == "$MP_DRIVER_NAME" ]; then
  if [ "$STD_DRIVER_INSTALLED" != "" ]; then
    # unload & uninstall std driver
    echo "A conflict has been discovered with a preexisting ATTO non-multipathing"
    echo "driver for this device."
    echo
    if [ "$INTERACTIVE" == "1" ]; then
        echo "You may uninstall the ${STD_DRIVER_NAME} driver here."
        echo "The driver module will be preserved and can be reinstalled if the ${DRIVER_NAME}"
        echo "driver is uninstalled using the uninstall script."
        echo
        echo "Would you like to unload and uninstall ${STD_DRIVER_NAME}?"
        read -s -n 1 RESPONSE
    else
        echo "Unloading and uninstalling ${STD_DRIVER_NAME}..."
        RESPONSE="y"
    fi
    if [ "${RESPONSE}" = "y" -o "${RESPONSE}" = "Y" ]; then
        # Unload the driver and rename the module so that it can possibly be reinstalled later
        rmmod ${STD_DRIVER_NAME}
        echo
        echo "${STD_DRIVER_NAME} unloaded."
        mv ${STD_DRIVER_INSTALLED} ${STD_DRIVER_INSTALLED/%.ko/.bkup}
        echo "${STD_DRIVER_NAME} backed up."
        echo
    fi
  fi
fi

if [ "$DRIVER_NAME" == "$MP_DRIVER_NAME" -o "$DRIVER_NAME" == "$STD_DRIVER_NAME" ]; then
  if [ "$CONFLICT_DRIVER_INSTALLED" != "" ]; then
    # unload & uninstall 3rd party driver conflict
    echo "A conflict has been discovered with a third-party driver (${CONFLICT_DRIVER_NAME})."
    echo
    if [ "$INTERACTIVE" == "1" ]; then
        echo "This could prevent the ATTO driver from operating properly."
        echo "You may uninstall the ${CONFLICT_DRIVER_NAME} driver here, although"
        echo "any devices that depend on it will no longer function."
        echo "The driver module will be preserved and can be reinstalled if the "
        echo "${DRIVER_NAME} driver is uninstalled using the uninstall script."
        echo
        echo "Would you like to unload and uninstall ${CONFLICT_DRIVER_NAME}?"
        read -s -n 1 RESPONSE
    else
        echo "Unloading and uninstalling ${CONFLICT_DRIVER_NAME}..."
        RESPONSE="y"
    fi
    if [ "${RESPONSE}" = "y" -o "${RESPONSE}" = "Y" ]; then
        # Unload the driver and rename the module so that it can possibly be reinstalled later
        rmmod ${CONFLICT_DRIVER_NAME}
        echo
        echo "${CONFLICT_DRIVER_NAME} unloaded."
        mv ${CONFLICT_DRIVER_INSTALLED} ${CONFLICT_DRIVER_INSTALLED/%.ko/.bkup}
        echo "${CONFLICT_DRIVER_NAME} backed up."
        echo
    fi
  fi
fi

STD_DRIVER_LOADED=`lsmod | grep ${STD_DRIVER_NAME} | grep -v ${MP_DRIVER_NAME}`
MP_DRIVER_LOADED=`lsmod | grep ${MP_DRIVER_NAME}`

# Build & install the modules

echo "Building and installing the ATTO ${PRETTY_NAME} driver, please wait..."

cd src

LOAD_MODULE="n"

if make -j clean install >make.log 2>make.err; then
    echo
    echo "Successfully installed the ${PRETTY_NAME} driver"
#    if lsmod | grep ${DRIVER_NAME} > /dev/null 2>/dev/null; then
    if [ "${DRIVER_NAME}" == "${STD_DRIVER_NAME}" -a "${STD_DRIVER_LOADED}" != "" ]; then
        echo
        echo "An older version of ${DRIVER_NAME} appears to be loaded"
        if [ "$INTERACTIVE" == "1" ]; then
            echo "Would you like to try to unload the old version and load the new one?  [N/y]"
            read -s -n 1 RESPONSE
        else
            echo "Unloading the old version and loading the new one..."
            RESPONSE="y"
            LOAD_MODULE="y"
        fi
        if [ "${RESPONSE}" = "y" -o "${RESPONSE}" = "Y" ]; then
            echo
            if ! rmmod ${DRIVER_NAME}; then
                echo "Sorry, could not unload the old driver"
                echo "You may need to reboot to use the new driver"
                REBOOT="Yes"
            else
                echo "Old version of ${DRIVER_NAME} successfully unloaded"
                LOAD_MODULE="y"
            fi
        fi
    elif [ "${DRIVER_NAME}" == "${MP_DRIVER_NAME}" -a "${MP_DRIVER_LOADED}" != "" ]; then
        echo
        echo "An older version of ${DRIVER_NAME} appears to be loaded"
        if [ "$INTERACTIVE" == "1" ]; then
            echo "Would you like to try to unload the old version and load the new one?  [N/y]"
            read -s -n 1 RESPONSE
        else
            echo "Unloading the old version and loading the new one..."
            RESPONSE="y"
            LOAD_MODULE="y"
        fi
        if [ "${RESPONSE}" = "y" -o "${RESPONSE}" = "Y" ]; then
            echo
            if ! rmmod ${DRIVER_NAME}; then
                echo "Sorry, could not unload the old driver"
                echo "You may need to reboot to use the new driver"
                REBOOT="Yes"
            else
                echo "Old version of ${DRIVER_NAME} successfully unloaded"
                LOAD_MODULE="y"
            fi
        fi
    else
        if [ "$INTERACTIVE" == "1" ]; then
            echo
            echo "Would you like to load the driver now?  [N/y]"
            read -s -n 1 LOAD_MODULE
        else
            LOAD_MODULE="y"
        fi
    fi
    echo
    if [ "${LOAD_MODULE}" = "y" -o "${LOAD_MODULE}" = "Y" ]; then
        echo "Please wait a moment while the driver loads..."
        echo
        if ! modprobe ${DRIVER_NAME}; then
            echo "Driver load failed!  Displaying relevant dmesg output:"
            echo "---------"
            dmesg | grep ${DRIVER_NAME} | tail -n 10
            echo "---------          "

            checkLoadStatus ${DRIVER_NAME}
        else
            echo "The ${DRIVER_NAME} driver has been loaded"

## TODO: rmmod the VNIC_DRIVER_NAME, check the load state and update REBOOT if necessary before modprobing the new one.

            if [ "${VNIC_DRIVER_NAME}" != "" ]; then
                if [ "$INTERACTIVE" == "1" ]; then
                    echo
                    echo "There is an additional driver (${VNIC_DRIVER_NAME}) to provide Ethernet access through your ${DRIVER_NAME} adapter"
                    echo
                    echo "Would you like to install the ${VNIC_DRIVER_NAME} virtual network driver as well [N/y]?"
                    read -s -n 1 LOAD_MODULE
                    echo
                else
                    LOAD_MODULE="y"
                fi

                if [ "${LOAD_MODULE}" = "y" -o "${LOAD_MODULE}" = "Y" ]; then
                    echo "Please wait a moment while the driver loads..."
                    echo
                    if ! modprobe ${VNIC_DRIVER_NAME}; then
                        echo "Driver load failed!  Displaying relevant dmesg output:"
                        echo "---------"
                        dmesg | grep ${VNIC_DRIVER_NAME} | tail -n 10
                        echo "---------          "

                        checkLoadStatus ${VNIC_DRIVER_NAME}
                    else
                        echo "The ${VNIC_DRIVER_NAME} driver has been loaded"
                    fi
                else
                    echo "You can load the driver using the command 'modprobe ${VNIC_DRIVER_NAME}'"
               fi

            fi
        fi
    else
        echo "You can load the driver using the command 'modprobe ${DRIVER_NAME}'"
    fi
    echo
else
    echo
    echo "An error occurred while making the driver"
    echo
    echo "---make output---"
    cat make.log
    echo
    echo "---make errors---"
    cat make.err
    echo "---------"
    echo
    pauseandexit 1
fi

# Execute the CLI tool installer - if it exists
cd ..
if [ -e ./cli_install.sh  ]; then
    ./cli_install.sh $1
fi

if [ $SHUTDOWN = "Yes" -a "$INTERACTIVE" = 0 ]; then
    pauseandexit 101
fi

if [ $REBOOT = "Yes" -a "$INTERACTIVE" = 0 ]; then
    pauseandexit 102
fi

pauseandexit 0
