


/****************************************************************************
;
; NAME: ostransport.c
;
;
; DESCRIPTION:
;
;       Linux driver FC transport interface
;
;
; REVISION HISTORY:
;
;       $Log: /ClrtyFC8/Linux/driver/ostransport.c $
; 
; 10    6/20/11 1:57p Pcullen
; Fix for IR7505 and associated kernel panics during driver removal.
; Also, instrumented Linux code with attolog statements.
; 
; 9     2/14/11 3:22p Pcullen
; Fixes to NPIV changes
; 
; 8     2/14/11 2:38p Pcullen
; Add support for NPIV in Celerity8 driver.
; 
; 7     4/22/10 8:55a Pcullen
; Comply with transport interface changes made for esas
; 
; 6     10/30/09 9:21a Pcullen
; VMware 4 porting
; Changed include to use oscommon.h.
; Removed legacy ESX 3.5 code - VMKERNEL_MODULE
; 
; 5     10/15/09 10:42a Jseba
; Removed all ATTO-specific data types
; 
; 4     10/02/09 3:18p Pcullen
; Fix compile error for fc transport in RH4
; 
; 3     8/06/09 1:29p Pcullen
; Provide more complete implementation of Linux FC Transport, including
; remote port api (rports)
; 
; 1     2/05/09 12:33p Pcullen
; Initial checkin from VMware 3.5 porting
; 
;
;
; THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
; ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
; WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
; PURSUANT TO WHICH IT WAS FURNISHED.
;
; COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2009
; ALL RIGHTS RESERVED.
;
;***************************************************************************/


#ifdef __VMKLNX__
#include "oscommon.h"
#else
#include "os.h"
#endif

#include "ostransport.h"

#ifdef ATTO_USE_TRANSPORT

#ifdef ATTO_TRANSPORT

#include <scsi/scsi_transport.h>
#include <scsi/scsi_transport_fc.h>

/* The FC transport originally (2.6.9) defined accessor functions called */
/* fc_node_name(), fc_port_name() fc_port_id(). These were replaced      */
/* in subsequent kernels with fc_starget_node_name(), ...                */
/* Since RH4 is based on 2.6.9, but contains backported changes, like    */
/* FC transport, we have to setup a compile-constant to determine        */
/* which set of macros to use. FC_TRANSPORT_USE_STARGET is that macro.   */
#undef FC_TRANSPORT_USE_STARGET
#if defined (fc_starget_node_name)
#define FC_TRANSPORT_USE_STARGET 1
#endif

#if LINUX_VERSION_CODE == KERNEL_VERSION(2,6,9) && !defined(FC_TRANSPORT_USE_STARGET)
/* The functions is this section are meant for a generic 2.6.9 kernel only */
static void celerityfc_get_node_name (struct scsi_device *sd)
{
    struct Scsi_Host *sh = sd->host;
    fc_node_name (sd) = OsTransport (sh->host_no,
                                     ATTO_TRANSPORT_TARG_WWNN,
                                     sd->id, NULL);
}

static void celerityfc_get_port_name (struct scsi_device *sd)
{
    struct Scsi_Host *sh = sd->host;
    fc_node_name (sd) = OsTransport (sh->host_no,
                                     ATTO_TRANSPORT_TARG_WWPN,
                                     sd->id, NULL);
}

static void celerityfc_get_port_id (struct scsi_device *sd)
{
    struct Scsi_Host *sh = sd->host;
    fc_node_name (sd) = OsTransport (sh->host_no,
                                     ATTO_TRANSPORT_TARG_PORTID,
                                     sd->id, NULL);
}
#endif /* LINUX_VERSION_CODE == KERNEL_VERSION(2,6,9) && */
       /* !defined(FC_TRANSPORT_USE_STARGET) */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10) || defined (FC_TRANSPORT_USE_STARGET)
/* The functions in this section are meant for at least kernel 2.6.10, or */
/* RedHat 4 or VMware 3.5                                                 */

static void celerityfc_get_starget_node_name (struct scsi_target *st)
{
    struct Scsi_Host *sh = dev_to_shost (st->dev.parent);

    fc_starget_node_name (st) = OsTransport (sh->host_no,
                                             ATTO_TRANSPORT_TARG_WWNN,
                                             st->id, NULL);
}

static void celerityfc_get_starget_port_name (struct scsi_target *st)
{
    struct Scsi_Host *sh = dev_to_shost (st->dev.parent);

    fc_starget_port_name (st) = OsTransport (sh->host_no,
                                             ATTO_TRANSPORT_TARG_WWPN,
                                             st->id, NULL);
}

static void celerityfc_get_starget_port_id (struct scsi_target *st)
{
    struct Scsi_Host *sh = dev_to_shost (st->dev.parent);

    fc_starget_port_id (st) = (u32) OsTransport (sh->host_no,
                                                 ATTO_TRANSPORT_TARG_PORTID,
                                                 st->id, NULL);
}
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10) ||           */
       /* defined (FC_TRANSPORT_USE_STARGET)                        */



#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
/* The functions in this section are meant for at least kernel 2.6.11 */

static void celerityfc_get_host_port_id (struct Scsi_Host *sh)
{
    fc_host_port_id (sh) = (u32) OsTransport (sh->host_no,
                                              ATTO_TRANSPORT_PORTID,
                                              0, NULL);
}

static void celerityfc_get_host_port_type (struct Scsi_Host *sh)
{
    u32 type;

    switch ((u32) OsTransport (sh->host_no, ATTO_TRANSPORT_PTYPE, 0, NULL))
    {
    case ATTO_TRANSPORT_PTYPE_LPORT:  type = FC_PORTTYPE_LPORT;  break;
    case ATTO_TRANSPORT_PTYPE_NLPORT: type = FC_PORTTYPE_NLPORT; break;
    case ATTO_TRANSPORT_PTYPE_PTP:    type = FC_PORTTYPE_PTP;    break;
    case ATTO_TRANSPORT_PTYPE_NPORT:  type = FC_PORTTYPE_NPORT;  break;
    default: type = FC_PORTTYPE_UNKNOWN;
    }

    fc_host_port_type (sh) = type;
}

static void celerityfc_get_host_port_state (struct Scsi_Host *sh)
{
    enum fc_port_state state;

    switch ((u32) OsTransport (sh->host_no, ATTO_TRANSPORT_PSTATE, 0, NULL))
    {
    case ATTO_TRANSPORT_PSTATE_ONLINE:  state = FC_PORTSTATE_ONLINE;  break;
    case ATTO_TRANSPORT_PSTATE_OFFLINE: state = FC_PORTSTATE_OFFLINE; break;
    default: state = FC_PORTSTATE_UNKNOWN;
    }

    fc_host_port_state (sh) = state;
}

static void celerityfc_get_host_speed (struct Scsi_Host *sh)
{
    fc_host_speed (sh) = (u32) OsTransport (sh->host_no,
                                            ATTO_TRANSPORT_SPEED,
                                            0, NULL);
}

static void celerityfc_get_host_fabric_name (struct Scsi_Host *sh)
{
    fc_host_fabric_name (sh) = OsTransport (sh->host_no,
                                            ATTO_TRANSPORT_FABRIC_NAME,
                                            0, NULL);
}

static struct fc_host_statistics *celerityfc_get_fc_host_stats (struct Scsi_Host *sh)
{
    struct fc_host_statistics *fhs = NULL;

    if (OsTransport (sh->host_no, ATTO_TRANSPORT_GETSTATS,
                     sizeof (struct fc_host_statistics), &fhs))
    {
        memset (fhs, -1, sizeof (struct fc_host_statistics));

        fhs->seconds_since_last_reset    = OsTransport (sh->host_no, ATTO_TRANSPORT_SSLR, 0, NULL);
        fhs->tx_frames                   = OsTransport (sh->host_no, ATTO_TRANSPORT_TXF_COUNT, 0, NULL);
        fhs->tx_words                    = OsTransport (sh->host_no, ATTO_TRANSPORT_TXW_COUNT, 0, NULL);
        fhs->rx_frames                   = OsTransport (sh->host_no, ATTO_TRANSPORT_RXF_COUNT, 0, NULL);
        fhs->rx_words                    = OsTransport (sh->host_no, ATTO_TRANSPORT_RXW_COUNT, 0, NULL);
        fhs->lip_count                   = OsTransport (sh->host_no, ATTO_TRANSPORT_LIP_COUNT, 0, NULL);
        fhs->nos_count                   = OsTransport (sh->host_no, ATTO_TRANSPORT_NOS_COUNT, 0, NULL);
        fhs->error_frames                = OsTransport (sh->host_no, ATTO_TRANSPORT_EF_COUNT, 0, NULL);
        fhs->dumped_frames               = OsTransport (sh->host_no, ATTO_TRANSPORT_DF_COUNT, 0, NULL);
        fhs->link_failure_count          = OsTransport (sh->host_no, ATTO_TRANSPORT_LF_COUNT, 0, NULL);
        fhs->loss_of_sync_count          = OsTransport (sh->host_no, ATTO_TRANSPORT_LSYNC_COUNT, 0, NULL);
        fhs->loss_of_signal_count        = OsTransport (sh->host_no, ATTO_TRANSPORT_LSIG_COUNT, 0, NULL);
        fhs->prim_seq_protocol_err_count = OsTransport (sh->host_no, ATTO_TRANSPORT_PSPE_COUNT, 0, NULL);
        fhs->invalid_tx_word_count       = OsTransport (sh->host_no, ATTO_TRANSPORT_ITXW_COUNT, 0, NULL);
        fhs->invalid_crc_count           = OsTransport (sh->host_no, ATTO_TRANSPORT_ICRC_COUNT, 0, NULL);
    }

    return fhs;
}

static void celerityfc_reset_fc_host_stats (struct Scsi_Host *sh)
{
    OsTransport (sh->host_no, ATTO_TRANSPORT_RESETSTATS,
                 sizeof (struct fc_host_statistics), NULL);
}

#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11) */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
static void celerityfc_set_dev_loss_tmo (struct fc_rport *rport, uint32_t timeout)
{
    if (timeout)
        rport->dev_loss_tmo = timeout;
    else
        rport->dev_loss_tmo = 1;
}
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12) */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
static int celerityfc_issue_fc_host_lip (struct Scsi_Host *sh)
{
    return OsTransport (sh->host_no, ATTO_TRANSPORT_LIP, 0, NULL);
}
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15) */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
static void celerityfc_get_host_symbolic_name (struct Scsi_Host *sh)
{
    OsTransport (sh->host_no,
                 ATTO_TRANSPORT_GET_SYM_NAME,
                 sizeof (fc_host_symbolic_name (sh)),
                 fc_host_symbolic_name (sh));
}
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19) */



#ifdef ATTO_USE_NPIV
static struct scsi_transport_template *celerityfc_vport_transport_template = NULL;

struct scsi_transport_template *atto_get_vport_transport_template (void)
{
	return celerityfc_vport_transport_template;
}




/****************************************************************************

    FUNCTION:   int celerityfc_vport_create

    PURPOSE:    Create a new virtual port based on NPIV.

    ARGUMENTS:  vport - pointer to the fc virtual port
                enable - flag to enable port after creation

    RETURNS: 0 - on success, negative error code on error

****************************************************************************/

int celerityfc_vport_create  (struct fc_vport *vport, bool disable)
{
    struct Scsi_Host *vhost, *phost;
    void **adapter_ctx;
    SHW shw;
    int error;
    int instance;
    int res = 0;
    int init_result=1, init_count;

    /* The parent (physical) scsi host */
    phost = vport->shost;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                  "create vport %llx:%llx disable:%d\n",
                  vport->node_name, vport->port_name, disable);

    /* This function roughly follows the adapter initialization found in */
    /* atto_init_device(), oswrap.c.                                     */

    vhost = scsi_host_alloc (&atto_vport_template, sizeof (void *));
    if (vhost == NULL)
    {
        ATTO_DEV_LOG (KERN_ERR, ATTO_LOG_NPIV, &(vport->dev),
                      "scsi_host_alloc() FAIL vport %llx:%llx\n",
                      vport->node_name, vport->port_name);
        return -ENODEV;
    }
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(vport->dev),
                  "scsi_host_alloc() OK %p\n", vhost);

    /* Override mas LUN and max target id */
    vhost->max_id  = ATTO_MAX_ID + 1;
    vhost->max_lun = ATTO_MAX_LUN + 1;

    /* We can handle 16-byte CDBs */
    vhost->max_cmd_len  = 16;

    vhost->can_queue    = can_queue;
    vhost->cmd_per_lun  = cmd_per_lun;
    vhost->this_id      = vhost->max_id + 1;
    vhost->max_channel  = 0;
    vhost->sg_tablesize = sg_tablesize;
    vhost->max_sectors  = atto_max_sectors;

    vhost->transportt = atto_get_vport_transport_template();

    adapter_ctx = (void **) vhost->hostdata;
    vport->dd_data = vhost;

#ifdef __VMKLNX__
    vport->vhost = vhost;
#endif

    ShwFromScsiHost (&shw, vhost);

    init_count = 1;

    while ( init_count <= npiv_init_retries )
    {
        init_result = OsInitVport (vport->node_name, vport->port_name,
                                   phost->hostdata, adapter_ctx,
                                   &shw, &instance);
        if (init_result)
        {
            break;
        }
        else 
        {
            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                          "Attempt #%d - failed to initialize vport %llx:%llx\n",
                          init_count, vport->node_name, vport->port_name);
            ++init_count;
            msleep (VPORT_INIT_POLL_TIME);
        }
    }

    if (!init_result)
    {
        ATTO_DEV_LOG (KERN_ERR, ATTO_LOG_NPIV, &(vport->dev),
                      "All attempts failed to initialize vport %llx:%llx\n",
                      vport->node_name, vport->port_name);

        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(vhost->shost_gendev),
                      "scsi_host_put() called\n");
        scsi_host_put (vhost);
        return -ENODEV;
    }


    vhost->unique_id = instance;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                  "virtual host instance is %d\n",
                  instance);

    error = scsi_add_host (vhost, &vport->dev);
    if (error)
    {
        ATTO_DEV_LOG (KERN_ERR, ATTO_LOG_NPIV, &(vhost->shost_gendev),
                      "scsi_add_host() FAIL vport %llx:%llx, error %d\n",
                      vport->node_name, vport->port_name, error);

        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(vhost->shost_gendev),
                      "scsi_host_put() called\n");
        scsi_host_put (vhost);
        return error;
    }
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(vhost->shost_gendev),
                  "scsi_add_host() OK %d\n", error);
    
    /* ATTO_USE_NPIV implies that ATTO_USE_TRANSPORT and */
    /* ATTO_TRANSPORT are defined, so it's OK to use     */
    /* use_transport_layer here without checking them.   */
    if (use_transport_layer)
        atto_init_host_attr (vhost, instance);

#if !defined (ATTO_RPORTS) && !defined (_MP_)
    /* Do not scan the host. In the case of FC Rports              */
    /* the devices are discovered via the transport layer as       */
    /* targets are reported. In the case of MP, the driver will    */
    /* report the devices as they are discovered.                  */
    scsi_scan_host (vhost);
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(vhost->shost_gendev),
                  "scsi_scan_host() called\n");
#else
    /* This is the case where RPORTS or MP is compiled in            */
    /* but the transport_layer is turned off by module param. In     */
    /* this case, we need to scan the host. See the comment above    */
    /* regarding use_transport_layer.                                */
    if (!use_transport_layer)
    {
        scsi_scan_host (vhost);
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(vhost->shost_gendev),
                      "scsi_scan_host() called\n");
    }

#endif

#ifdef ATTO_RESCANS_ENABLED
#if defined (ATTO_RPORTS)
    /* If using FC rports then process all scan state */
    /* changes.                                       */
    if (use_transport_layer)
        OsAllowChangeNotification (instance, 1, 0);
    else
        OsAllowChangeNotification (instance, 1, 1);

#else
    /* If not using FC rports then clear the */
    /* ScanState table (third param = 1).    */
    OsAllowChangeNotification (instance, 1, 1);
#endif

#if defined (ATTO_RPORTS)
    if (use_transport_layer)
        LinuxScheduleWork (instance);
#endif
#endif /* ATTO_RESCANS_ENABLED */

    if (!disable)
    {
        unsigned long disc_wait_time;
        unsigned long disc_start_time;

        disc_wait_time = msecs_to_jiffies(npiv_disc_wait_time);
        disc_wait_time += jiffies;
        disc_start_time = jiffies;

        /* enable the vport */
        res = celerityfc_vport_disable (vport, disable);

        /* Wait until really ready */

        while ( time_before(jiffies, disc_wait_time) )
        {
            if ( OsVportReady(vhost->hostdata) )
            {
                ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                              "vport ready after %dms\n",
                              jiffies_to_msecs(jiffies - disc_start_time));
                break;
            }
            msleep(VPORT_DISC_POLL_TIME);
        }
        if (jiffies > disc_wait_time)
        {
            ATTO_DEV_LOG (KERN_WARNING, ATTO_LOG_NPIV, &(vport->dev),
                          "vport not ready after %dms\n; Gave up.",
                          jiffies_to_msecs(jiffies - disc_start_time));
        }
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                  "created vport WWN %llx:%llx\n",
                  vport->node_name, vport->port_name);

    return res;
}

int celerityfc_vport_disable (struct fc_vport *vport , bool disable)
{
    int res;
    struct Scsi_Host *vhost = (struct Scsi_Host *)vport->dd_data;
    void *ctx;

    if (disable)
    {
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                      "disabling vport %llx:%llx\n",
                      vport->node_name, vport->port_name);
    }
    else
    {
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                      "enabling vport %llx:%llx\n",
                      vport->node_name, vport->port_name);
    }


    if (vhost == NULL)
    {
        ATTO_DEV_LOG (KERN_WARNING, ATTO_LOG_NPIV, &(vport->dev),
                      "NULL scsi host for vport %llx:%llx\n",
                      vport->node_name, vport->port_name);
        return -EINVAL;
    }

    ctx = vhost->hostdata;
    res = OsDisableVport (ctx, disable);

    if (res)
    {
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                      "%s\n", disable ? "disabled" : "enabled");
        return 0;
    }
    else
    {
        ATTO_DEV_LOG (KERN_WARNING, ATTO_LOG_NPIV, &(vport->dev),
                      "failed to %s vport\n",
                      disable ? "disable" : "enable");
        return -EINVAL;
        
    }
}

int celerityfc_vport_delete  (struct fc_vport *vport)
{
    int res;
    struct Scsi_Host *shost = (struct Scsi_Host *) vport->shost;
    struct Scsi_Host *vhost = (struct Scsi_Host *) vport->dd_data;
    void *ctx;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                  "delete vport %llx:%llx\n",
                  vport->node_name, vport->port_name);

    if (vhost == NULL)
    {
        ATTO_DEV_LOG (KERN_WARNING, ATTO_LOG_NPIV, &(vport->dev),
                      "NULL scsi host for vport %llx:%llx\n",
                      vport->node_name, vport->port_name);
        return -EINVAL;
    }

    if (vport->vport_state != FC_VPORT_DISABLED)
    {
        unsigned long disable_wait_time;
        unsigned long disable_start_time;

        disable_wait_time = msecs_to_jiffies(npiv_disc_wait_time);
        disable_wait_time += jiffies;
        disable_start_time = jiffies;

        /* Disable the vport */
        res = celerityfc_vport_disable (vport, 1);

        /* Wait until really disabled */

        while ( time_before(jiffies, disable_wait_time) )
        {
            if ( vport->vport_state == FC_VPORT_DISABLED )
            {
                ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                              "vport disabled after %dms\n",
                              jiffies_to_msecs(jiffies - disable_start_time));
                break;
            }
            msleep(VPORT_DISC_POLL_TIME);
        }
        if (jiffies > disable_wait_time)
        {
            ATTO_DEV_LOG (KERN_WARNING, ATTO_LOG_NPIV, &(vport->dev),
                          "vport not disabled after %dms\n; Gave up.",
                          jiffies_to_msecs(jiffies - disable_start_time));
        }
    }


    ctx = vhost->hostdata;
    res = OsDeleteVport (ctx);

    if (res)
    {
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_NPIV, &(vport->dev),
                      "deleted vport WWN %llx:%llx\n",
                      vport->node_name, vport->port_name);

        /* Remove the virtual fc host and scsi host */
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(vhost->shost_gendev),
                      "fc_remove_host() called\n");
        fc_remove_host (vhost);

        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(vhost->shost_gendev),
                      "scsi_remove_host() called\n");
        scsi_remove_host (vhost);

        /* Remove our reference to the virtual host */
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(vhost->shost_gendev),
                      "scsi_host_put() called\n");
        scsi_host_put (vhost);

        /* Remove our reference to the physical scsi host */
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(shost->shost_gendev),
                      "scsi_host_put() called\n");
        scsi_host_put (shost);
        return 0;
    }
    else
    {
        ATTO_DEV_LOG (KERN_WARNING, ATTO_LOG_NPIV, &(vport->dev),
                      "failed to delete vport\n");
        return -EINVAL;
    }

}

#endif /* ATTO_USE_NPIV */





struct fc_function_template celerityfc_transport_functions =
{
#if LINUX_VERSION_CODE == KERNEL_VERSION(2,6,9) && !defined(FC_TRANSPORT_USE_STARGET)
/* Generic 2.6.9 kernel */
    .get_port_id                   = celerityfc_get_port_id,
    .get_node_name                 = celerityfc_get_node_name,
    .get_port_name                 = celerityfc_get_port_name,

    .show_port_id                  = 1,
    .show_node_name                = 1,
    .show_port_name                = 1,
#endif /* LINUX_VERSION_CODE == KERNEL_VERSION(2,6,9) && !defined(FC_TRANSPORT_USE_STARGET) */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10) || defined (FC_TRANSPORT_USE_STARGET)
/* Generic 2.6.10, or RH4 or VMware 3.5 */
    .get_starget_node_name         = celerityfc_get_starget_node_name,
    .get_starget_port_name         = celerityfc_get_starget_port_name,
    .get_starget_port_id           = celerityfc_get_starget_port_id,

    .show_starget_node_name        = 1,
    .show_starget_port_name        = 1,
    .show_starget_port_id          = 1,
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10) ||     */
       /* defined (FC_TRANSPORT_USE_STARGET)                  */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
/* Generic 2.6.11 */

    .get_host_port_id              = celerityfc_get_host_port_id,
    .get_host_port_type            = celerityfc_get_host_port_type,
    .get_host_port_state           = celerityfc_get_host_port_state,
    .get_host_speed                = celerityfc_get_host_speed,
    .get_host_fabric_name          = celerityfc_get_host_fabric_name,
    .get_fc_host_stats             = celerityfc_get_fc_host_stats,
    .reset_fc_host_stats           = celerityfc_reset_fc_host_stats,

    /* host fixed attributes */
    .show_host_node_name           = 1,
    .show_host_port_name           = 1,
    .show_host_supported_classes   = 1,
    .show_host_supported_fc4s      = 1,
    .show_host_supported_speeds    = 1,
    .show_host_maxframe_size       = 1,

    /* host dynamic attributes */
    .show_host_port_id             = 1,
    .show_host_port_type           = 1,
    .show_host_port_state          = 1,
    .show_host_speed               = 1,
    .show_host_fabric_name         = 1,
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11) */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,12)
    /* allocation lengths for host-specific data */
    .dd_fcrport_size               = sizeof (unsigned int),

    /* remote port fixed attributes */
    .show_rport_maxframe_size      = 1,
    .show_rport_supported_classes  = 1,
    .show_rport_dev_loss_tmo       = 1,
    .set_rport_dev_loss_tmo        = celerityfc_set_dev_loss_tmo,
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
    .issue_fc_host_lip             = celerityfc_issue_fc_host_lip,
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15) */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
    .show_host_permanent_port_name = 0,
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16) */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,17)
    .show_rport_maxframe_size      = 1,
    .show_rport_supported_classes  = 1,
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
    .get_host_symbolic_name        = celerityfc_get_host_symbolic_name,
    .dev_loss_tmo_callbk           = NULL,
    .terminate_rport_io            = NULL,
    .set_host_system_hostname      = NULL,
    .show_host_system_hostname     = 1,
    .show_host_symbolic_name       = 1,
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19) */


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
    .tsk_mgmt_response             = NULL,
    .it_nexus_response             = NULL,
    .disable_target_scan           = 0,
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25) */

#ifdef ATTO_USE_NPIV
    .vport_create                  = celerityfc_vport_create,
    .vport_disable                 = celerityfc_vport_disable,
    .vport_delete                  = celerityfc_vport_delete,
#endif /* ATTO_USE_NPIV */

};

#ifdef ATTO_USE_NPIV
struct fc_function_template celerityfc_vport_transport_functions =
{

    .get_starget_node_name         = celerityfc_get_starget_node_name,
    .get_starget_port_name         = celerityfc_get_starget_port_name,
    .get_starget_port_id           = celerityfc_get_starget_port_id,

    .show_starget_node_name        = 1,
    .show_starget_port_name        = 1,
    .show_starget_port_id          = 1,

    .get_host_port_id              = celerityfc_get_host_port_id,
    .get_host_port_type            = celerityfc_get_host_port_type,
    .get_host_port_state           = celerityfc_get_host_port_state,
    .get_host_speed                = celerityfc_get_host_speed,
    .get_host_fabric_name          = celerityfc_get_host_fabric_name,
    .get_fc_host_stats             = celerityfc_get_fc_host_stats,
    .reset_fc_host_stats           = celerityfc_reset_fc_host_stats,

    /* host fixed attributes */
    .show_host_node_name           = 1,
    .show_host_port_name           = 1,
    .show_host_supported_classes   = 1,
    .show_host_supported_fc4s      = 1,
    .show_host_supported_speeds    = 1,
    .show_host_maxframe_size       = 1,

    /* host dynamic attributes */
    .show_host_port_id             = 1,
    .show_host_port_type           = 1,
    .show_host_port_state          = 1,
    .show_host_speed               = 1,
    .show_host_fabric_name         = 1,


    /* allocation lengths for host-specific data */
    .dd_fcrport_size               = sizeof (unsigned int),

    /* remote port fixed attributes */
    .show_rport_maxframe_size      = 1,
    .show_rport_supported_classes  = 1,
    .show_rport_dev_loss_tmo       = 1,
    .set_rport_dev_loss_tmo        = celerityfc_set_dev_loss_tmo,


    .issue_fc_host_lip             = celerityfc_issue_fc_host_lip,

    .show_host_permanent_port_name = 0,

    .show_rport_maxframe_size      = 1,
    .show_rport_supported_classes  = 1,


    .get_host_symbolic_name        = celerityfc_get_host_symbolic_name,
    .dev_loss_tmo_callbk           = NULL,
    .terminate_rport_io            = NULL,
    .set_host_system_hostname      = NULL,
    .show_host_system_hostname     = 1,
    .show_host_symbolic_name       = 1,


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
    .tsk_mgmt_response             = NULL,
    .it_nexus_response             = NULL,
    .disable_target_scan           = 0,
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25) */

    .vport_disable                 = celerityfc_vport_disable,

};
#endif /* ATTO_USE_NPIV */

/*************************************************************************/
/* Public Functions                                                      */
/* These functions are used for attaching the transport, releasing the   */
/* transport and initializing the transport variables. These functions   */
/* are meant to be portable across all kernel versions.                  */
/*************************************************************************/

static struct scsi_transport_template *celerityfc_transport_template = NULL;

struct scsi_transport_template *atto_get_transport_template (void)
{
    return celerityfc_transport_template;
}


int atto_attach_transport (void)
{

    celerityfc_transport_template = fc_attach_transport (&celerityfc_transport_functions);

    if (celerityfc_transport_template == NULL)
    {
        ATTO_LOG (KERN_WARNING, ATTO_LOG_XPORT, "fc_attach_transport(host) FAIL\n");
        return 0;
    }
    ATTO_LOG (KERN_INFO, ATTO_LOG_XPORT, "fc_attach_transport(host) OK\n");

#ifdef ATTO_USE_NPIV
    celerityfc_vport_transport_template = fc_attach_transport (&celerityfc_vport_transport_functions);

	if (celerityfc_vport_transport_template == NULL)
        {
            ATTO_LOG (KERN_WARNING, ATTO_LOG_XPORT, "fc_attach_transport(vport) FAIL\n"); 
            return 0;
        }
        ATTO_LOG (KERN_INFO, ATTO_LOG_XPORT, "fc_attach_transport(vport) OK\n");
#endif /* ATTO_USE_NPIV */


    return 1;
}


void atto_release_transport (void)
{
    if (celerityfc_transport_template)
    {
        ATTO_LOG (KERN_INFO, ATTO_LOG_XPORT, "fc_release_transport(host) called\n");
        fc_release_transport (celerityfc_transport_template);
        celerityfc_transport_template = NULL;
    }

#ifdef ATTO_USE_NPIV
    if (celerityfc_vport_transport_template)
    {
        ATTO_LOG (KERN_INFO, ATTO_LOG_XPORT, "fc_release_transport(vport) called\n");
    	fc_release_transport (celerityfc_vport_transport_template);
    	celerityfc_vport_transport_template = NULL;
    }
#endif /* ATTO_USE_NPIV */
}


/* Initialize the static attributes for the FC transport */
void atto_init_host_attr (struct Scsi_Host *sh, int index)
{
    if (sh)
    {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11)
        /* host_supported_classes */
        unsigned int class_bits = OsTransport (sh->host_no, ATTO_TRANSPORT_CLASS, 0, NULL);
        unsigned int speed_bits = OsTransport (sh->host_no, ATTO_TRANSPORT_SSPEED, 0, NULL);

        fc_host_supported_classes (sh) = 0;
        if (class_bits & 0x08) fc_host_supported_classes (sh) |= FC_COS_CLASS3;
        if (class_bits & 0x04) fc_host_supported_classes (sh) |= FC_COS_CLASS2;
        if (class_bits & 0x02) fc_host_supported_classes (sh) |= FC_COS_CLASS1;

        /* host_supported_speeds */
        fc_host_supported_speeds (sh) = 0;
#ifdef FC_PORTSPEED_16GBIT
        if (speed_bits & (1 << 16)) fc_host_supported_speeds (sh) |= FC_PORTSPEED_10GBIT;
#endif
#ifdef FC_PORTSPEED_10GBIT
        if (speed_bits & (1 << 10)) fc_host_supported_speeds (sh) |= FC_PORTSPEED_10GBIT;
#endif
#ifdef FC_PORTSPEED_8GBIT
        if (speed_bits & (1 << 8))  fc_host_supported_speeds (sh) |= FC_PORTSPEED_8GBIT;
#endif
        if (speed_bits & (1 << 4))  fc_host_supported_speeds (sh) |= FC_PORTSPEED_4GBIT;
        if (speed_bits & (1 << 2))  fc_host_supported_speeds (sh) |= FC_PORTSPEED_2GBIT;
        if (speed_bits & (1 << 2))  fc_host_supported_speeds (sh) |= FC_PORTSPEED_1GBIT;

        /* host_node_name */
        fc_host_node_name (sh) = OsTransport (sh->host_no, ATTO_TRANSPORT_WWNN, 0, NULL);

        /* host_port_name */
        fc_host_port_name (sh) = OsTransport (sh->host_no, ATTO_TRANSPORT_WWPN, 0, NULL);

        /* host_maxframe_size */
        fc_host_maxframe_size (sh) = (u32) OsTransport (sh->host_no, ATTO_TRANSPORT_FRAME_SIZE, 0, NULL);

        /* host_supported_fc4s */
        OsTransport (sh->host_no, ATTO_TRANSPORT_SUPPORTED_FC4, 
                     sizeof (fc_host_supported_fc4s (sh)), fc_host_supported_fc4s (sh));

        /* host_active_fc4s */
        OsTransport (sh->host_no, ATTO_TRANSPORT_ACTIVE_FC4, 
                     sizeof (fc_host_active_fc4s (sh)), fc_host_active_fc4s (sh));

#ifdef ATTO_USE_NPIV
        fc_host_max_npiv_vports(sh) = max_npiv_vports;
#endif /* ATTO_USE_NPIV */
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,11) */

    }
    else
        ATTO_LOG (KERN_WARNING, ATTO_LOG_RPORT, 
                  "scsi host is null in atto_init_host_attr\n");
}


#ifdef ATTO_RPORTS

static struct fc_rport * find_rport_from_target_id (struct Scsi_Host *host, unsigned int target_id)
{
    struct fc_rport *rport;
    unsigned long flags;
    struct fc_host_attrs *fc_host = shost_to_fc_host (host);

    /* Search the list of active rports on the scsi host */
    spin_lock_irqsave (host->host_lock, flags);
    list_for_each_entry (rport, &fc_host->rports, peers)
    {
        if (*((unsigned int *)rport->dd_data) == target_id)
        {
            spin_unlock_irqrestore (host->host_lock, flags);
            return rport;
        }
    }

    spin_unlock_irqrestore (host->host_lock, flags);
    return NULL;
}

void celerityfc_add_rport (struct Scsi_Host *host, unsigned int target_id)
{
    struct fc_rport *rport;
    struct fc_rport_identifiers rport_ids;

    /* Get rport identifiers via OsTransport() calls */
    rport_ids.node_name = OsTransport (host->host_no, ATTO_TRANSPORT_TARG_WWNN,
                                       target_id, NULL);
    rport_ids.port_name = OsTransport (host->host_no, ATTO_TRANSPORT_TARG_WWPN,
                                       target_id, NULL);
    rport_ids.port_id = OsTransport (host->host_no, ATTO_TRANSPORT_TARG_PORTID,
                                     target_id, NULL);
    rport_ids.roles = FC_RPORT_ROLE_FCP_TARGET;

    /* Call transport-layer to add rport */
    rport = fc_remote_port_add (host, 0, &rport_ids);
    if (rport == NULL) 
    {
        ATTO_DEV_LOG (KERN_ERR, ATTO_LOG_RPORT, &(host->shost_gendev),
                  "fc_remote_port_add() FAIL\n");
        return;
    }
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_RPORT, &(rport->dev),
                  "fc_remote_port_add() OK %p\n", rport);
    
    /* Set addnl rport info: dd_data=target_id, role, ... */
    memcpy((unsigned long *)rport->dd_data, &target_id, sizeof (unsigned long));
    rport->maxframe_size = fc_host_maxframe_size (host);
    rport->supported_classes = fc_host_supported_classes (host);

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_RPORT, &(rport->dev),
                  "created rport WWN:%llx:%llx T:%d state:%d\n",
                  rport_ids.node_name, rport_ids.port_name, target_id,
                  rport->port_state);
}

void celerityfc_remove_rport (struct Scsi_Host *host, unsigned int target_id)
{
    struct fc_rport *rport;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_RPORT, &(host->shost_gendev),
                  "removing rport at target %d\n", target_id);

    /* Find rport where dd_data==target_id from those stored in host */
    rport = find_rport_from_target_id (host, target_id);
    if (rport == NULL)
    {
        ATTO_DEV_LOG (KERN_WARNING, ATTO_LOG_RPORT, &(host->shost_gendev),
                      "no rport at target %d\n",
                      target_id);
        return;
    }

    /* Call tranport-layer to remove rport */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(rport->dev),
                  "fc_remote_port_delete() called; target %d\n",
                  target_id);
    fc_remote_port_delete (rport);

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_RPORT, &(host->shost_gendev),
                  "removed rport at target %d\n", target_id);
}

void celerityfc_scan_rport (struct Scsi_Host *host, unsigned int target_id)
{
    struct fc_rport *rport;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_RPORT, &(host->shost_gendev),
                  "scanning rport at target %d\n", target_id);

    /* Find rport where dd_data==target_id from those stored in host */
    rport = find_rport_from_target_id (host, target_id);
    if (rport == NULL)
    {
        ATTO_DEV_LOG (KERN_WARNING, ATTO_LOG_RPORT, &(host->shost_gendev),
                      "no rport at target %d\n", target_id);
        return;
    }

    /* Call mid-layer to invoke a scan on rport??? */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, &(rport->dev),
                  "scsi_scan_target() called; scsi_target_id %d\n",
                  rport->scsi_target_id);
    scsi_scan_target (&rport->dev, rport->channel, rport->scsi_target_id,
                      SCAN_WILD_CARD, 1);
}

#endif /* ATTO_RPORTS */


#endif /* ATTO_TRANSPORT */

#endif /* ATTO_USE_TRANSPORT */
